import { PrismaClient } from '@prisma/client'
import { CreateSessionData, Session } from 'src/modules/auth/models'
import { AppError } from 'src/packages/fastify-compliant-http-errors'
import {Uuid} from '@theoryofnekomata/uuid-buffer';
import {notFoundFactory} from 'src/packages/prisma-error-utils';
import {User} from '../user';

type SessionWithUsername = Session & { user: Pick<User, 'username'> };

export interface SessionService {
	create(data: CreateSessionData): Promise<SessionWithUsername>

	expire(sessionId: Session['id']): Promise<Session>

	isValid(sessionId: Session['id']): Promise<boolean>

	get(id: Session['id']): Promise<SessionWithUsername | null>
}

export class SessionNotFoundError extends AppError {}

export class SessionServiceImpl implements SessionService {
	private readonly prismaClient: PrismaClient

	constructor() {
		this.prismaClient = new PrismaClient()
	}

	private static serialize<T extends Session>(raw: T): T {
		return {
			...raw,
			id: Uuid.from(raw.id),
			userId: Uuid.from(raw.userId)
		};
	}

	async create(data: CreateSessionData): Promise<SessionWithUsername> {
		const createdAt = new Date()
		const validUntil = new Date(createdAt)
		validUntil.setDate(validUntil.getDate() + 1)
		const raw = await this.prismaClient.session.create({
			data: {
				...data,
				id: Uuid.v4(),
				createdAt,
				validUntil,
			},
			include: {
				user: {
					select: {
						id: false,
						username: true,
						password: false,
					},
				},
			},
		})

		return SessionServiceImpl.serialize(raw);
	}

	async expire(sessionId: Session['id']): Promise<Session> {
		const existingSession = await this.prismaClient.session.findUnique({
			where: {
				id: sessionId,
			},
			rejectOnNotFound: notFoundFactory(SessionNotFoundError),
		})
		const raw = await this.prismaClient.session.update({
			where: {
				id: existingSession.id,
			},
			data: {
				validUntil: new Date(),
			},
		})

		return SessionServiceImpl.serialize(raw);
	}

	async isValid(sessionId: Session['id']): Promise<boolean> {
		const data = await this.prismaClient.session.findUnique({
			where: {
				id: sessionId,
			},
		})

		if (!data) {
			return false
		}
		const now = new Date()
		return (
			data.createdAt <= now
			&& now < data.validUntil
		)
	}

	async get(id: Session['id']): Promise<SessionWithUsername | null> {
		const raw = await this.prismaClient.session.findUnique({
			where: {
				id,
			},
			include: {
				user: {
					select: {
						id: false,
						username: true,
						password: false,
					}
				}
			}
		})

		if (raw) {
			return SessionServiceImpl.serialize(raw);
		}

		return null;
	}
}
