import {Uuid} from '@theoryofnekomata/uuid-buffer';
import {OwnerType} from './models';
import {UnknownOwnerTypeError} from './responses';
import {UserNotFoundError, UserService, UserServiceImpl} from '../user';
import {OrgNotFoundError, OrgService, OrgServiceImpl} from '../org';

export interface OwnerService {
  getOwnerName(id: Uuid, type: string): Promise<string>;
  getOwnerTypeDirName(type: string): string;
}

export class OwnerServiceImpl implements OwnerService {
  private readonly userService: UserService;
  private readonly orgService: OrgService;
  private readonly OWNER_TYPE_DIR_NAME: Record<OwnerType, string>;

  constructor() {
    this.userService = new UserServiceImpl();
    this.orgService = new OrgServiceImpl();
    this.OWNER_TYPE_DIR_NAME = {
      [OwnerType.USER]: 'users',
      [OwnerType.ORG]: 'orgs',
    };
  }

  private async getOwnerNameFromUser(ownerId: Uuid) {
    try {
      const user = await this.userService.getFromId(ownerId);
      return user.username;
    } catch (err) {
      if (!(err instanceof UserNotFoundError)) {
        throw err;
      }
    }

    return ownerId.toString();
  }

  private async getOwnerNameFromOrg(ownerId: Uuid) {
    try {
      const org = await this.orgService.getFromId(ownerId);
      return org.name;
    } catch (err) {
      if (!(err instanceof OrgNotFoundError)) {
        throw err;
      }
    }

    return ownerId.toString();
  }

  async getOwnerName(ownerId: Uuid, ownerType: string) {
    if (ownerType === OwnerType.USER) {
      return this.getOwnerNameFromUser(ownerId);
    }

    if (ownerType === OwnerType.ORG) {
      return this.getOwnerNameFromOrg(ownerId);
    }

    throw new UnknownOwnerTypeError(`Only valid options are: ${Object.values(OwnerType).join(', ')}`);
  }

  getOwnerTypeDirName(type: string): string {
    const { [type as OwnerType]: dirName } = this.OWNER_TYPE_DIR_NAME;

    if (!dirName) {
      throw new UnknownOwnerTypeError(`Only valid options are: ${Object.values(OwnerType).join(', ')}`);
    }

    return dirName;
  }
}
