import {RepoService, RepoServiceImpl} from './Repo.service';
import {RouteHandlerMethod} from 'fastify';
import {constants} from 'http2';
import {HttpResponse} from '../../packages/fastify-send-data';
import { Controller } from 'src/packages/fastify-utils-theoryofnekomata'
import {Uuid} from '@theoryofnekomata/uuid-buffer';
import {CreateRepoData, Repo} from './models';

export interface RepoController extends Controller<
  'createRepo'
  | 'deleteRepo'
  | 'getRepoRefs'
  | 'receivePack'
> {}

class RepoCreatedResponse extends HttpResponse<Repo>(constants.HTTP_STATUS_CREATED, 'Repo Created') {}
class RepoDeletedResponse extends HttpResponse(constants.HTTP_STATUS_NO_CONTENT, 'Repo Deleted') {}

export class RepoControllerImpl implements RepoController {
  private readonly repoService: RepoService;
  private readonly noCacheHeaders = {
    'Expires': 'Fri, 01 Jan 1980 00:00:00 GMT',
    'Pragma': 'no-cache',
    'Cache-Control': 'no-cache, max-age=0, must-revalidate',
  };

  constructor() {
    this.repoService = new RepoServiceImpl();
  }

  readonly createRepo: RouteHandlerMethod = async (request, reply) => {
    const repo = await this.repoService.createRepo(request.body as CreateRepoData, request.session?.user);
    reply.sendData(new RepoCreatedResponse(repo));
  }

  readonly deleteRepo: RouteHandlerMethod = async (request, reply) => {
    const params = request.params as { repoId: string }
    await this.repoService.deleteRepo(Uuid.from(params.repoId), request.session?.user)
    reply.sendData(new RepoDeletedResponse());
  }

  readonly getRepoRefs: RouteHandlerMethod = async (request, reply) => {
    const { service } = request.query as { service: string };
    const params = request.params as { repoName: string, ownerName: string, ownerType: string }
    const outputStream = await this.repoService.getRefsFromService(
      params.ownerType,
      params.ownerName,
      params.repoName,
      service,
      request.session?.user,
    );

    reply
      .headers({
        ...this.noCacheHeaders,
        'Content-Type': `application/x-${service}-advertisement`,
      })
      .send(outputStream);
  }

  readonly receivePack: RouteHandlerMethod = async (request, reply) => {
    console.log('content-type', request.headers['content-type']);
    console.log('body', typeof request.body, request.body, (request.body as Buffer).toString('utf-8'));
    reply
      .headers({
        ...this.noCacheHeaders,
        // Mark this as service response
        'Content-Type': 'application/x-git-receive-pack-result',
      })
      .status(500)
      .send('');
  }
}
