import {PrismaClient} from '@prisma/client';
import {PassThrough, Stream} from 'stream';
import {Uuid} from '@theoryofnekomata/uuid-buffer';
import * as codeCore from '@modal/code-core';
import {notFoundFactory} from '../../packages/prisma-error-utils';
import {User} from '../user';
import {CreateRepoData, Repo, RepoAvailableService} from './models';
import {
  UnauthorizedToCreateRepoError,
  UnauthorizedToDeleteRepoError,
  UnauthorizedToInvokeRepoServiceError,
  UnableToCreateRepoError,
  UnableToDeleteRepoError,
  UnableToInvokeRepoServiceError,
  UnknownServiceError,
  RepoNotFoundError,
} from './responses';
import {ChildProcess} from 'child_process';
import path from 'path';
import {OwnerService, OwnerServiceImpl, OwnerType} from '../owner';
import {RepoLogService, RepoLogServiceImpl} from '../repo-log';

type GitServiceMethod = (cwd: string) => Promise<ChildProcess>;

export interface RepoService {
  createRepo(data: CreateRepoData, user?: User): Promise<Repo>;
  getSingleRepoById(repoId: Repo['id']): Promise<Repo>;
  getSingleRepoByOwnerAndName(ownerType: Repo['ownerType'], ownerId: Repo['ownerId'], repoName: Repo['name']): Promise<Repo>;
  getOwnedRepos(ownerType: Repo['ownerType'], ownerId: Repo['ownerId']): Promise<Repo[]>;
  deleteRepo(repoId: Repo['id'], user?: User): Promise<void>;
  getRefsFromService(
    ownerType: Repo['ownerType'],
    ownerName: Repo['name'],
    repoName: Repo['name'],
    service?: string,
    user?: User,
  ): Promise<Stream>;
}

export class RepoServiceImpl implements RepoService {
  private readonly gitService: codeCore.git.GitService;
  private readonly repoLogService: RepoLogService;
  private readonly prismaClient: PrismaClient;
  private readonly ownerService: OwnerService;

  constructor() {
    this.gitService = new codeCore.git.GitServiceImpl();
    this.repoLogService = new RepoLogServiceImpl();
    this.prismaClient = new PrismaClient();
    this.ownerService = new OwnerServiceImpl();
  }
  
  private static serialize<T extends Repo>(repoMetadataRaw: T): T {
    return {
      ...repoMetadataRaw,
      id: Uuid.from(repoMetadataRaw.id),
      ownerId: Uuid.from(repoMetadataRaw.ownerId),
    };
  }

  private static getRepoBasePath(ownerType: string, ownerName: string, repoName: string) {
    return path.join('repos', ownerType, ownerName, repoName.endsWith('.git') ? repoName : `${repoName}.git`);
  }

  private static authenticate(_user?: User): User {
    return {
      id: Uuid.v4(),
      username: 'foo',
    } as User;
  }

  private async doCreateRepoFilesProcess(repoName: string, ownerName: string, ownerType: OwnerType): Promise<ChildProcess> {
    try {
      const ownerTypeDirName = this.ownerService.getOwnerTypeDirName(ownerType);
      const repoBasePath = RepoServiceImpl.getRepoBasePath(repoName, ownerName, ownerTypeDirName);
      return this.gitService.create(repoBasePath);
    } catch (causeRaw) {
      const cause = causeRaw as Error;
      throw new UnableToCreateRepoError('Something went wrong while creating the repo.', { cause, });
    }
  }

  private async doCreateRepoFiles(repoName: string, ownerName: string, ownerType: OwnerType): Promise<void> {
    const childProcess = await this.doCreateRepoFilesProcess(repoName, ownerName, ownerType);
    return new Promise((resolve, reject) => {
      childProcess.on('error', (causeRaw) => {
        const cause = causeRaw as Error;
        reject(new UnableToCreateRepoError('Something went wrong while creating the repo.', { cause, }));
      });

      childProcess.on('close', (code) => {
        if (code !== 0) {
          reject(new UnableToCreateRepoError('Something went wrong while creating the repo.', {
            cause: new Error(`Git process returned ${code}`),
          }));
          return;
        }
        resolve();
      });
    })
  }

  private async doCreateRepoMetadata(repoId: Repo['id'], data: CreateRepoData): Promise<Repo> {
    let repoMetadataRaw: Repo;
    try {
      repoMetadataRaw = await this.prismaClient.repo.create({
        data: {
          id: repoId,
          name: data.name,
          visibility: data.visibility,
          ownerId: data.ownerId,
          ownerType: data.ownerType,
        },
      });
    } catch (causeRaw) {
      const cause = causeRaw as Error;
      throw new UnableToCreateRepoError('Something went wrong while creating the repo.', { cause, });
    }

    return RepoServiceImpl.serialize(repoMetadataRaw);
  }

  async createRepo(data: CreateRepoData, userRaw?: User): Promise<Repo> {
    // TODO put this in ability service
    const user = RepoServiceImpl.authenticate(userRaw);
    if (!user) {
      throw new UnauthorizedToCreateRepoError('Could not create repo with insufficient authorization.');
    }

    const repoId = Uuid.v4();
    const ownerId = Uuid.from(data.ownerId);
    const ownerName = await this.ownerService.getOwnerName(ownerId, data.ownerType);
    await this.doCreateRepoFiles(data.name, ownerName, data.ownerType as OwnerType);
    const repoMetadata = await this.doCreateRepoMetadata(repoId, {
      ...data,
      ownerId,
    });

    await this.repoLogService.logCreateRepo(user, repoMetadata, ownerName);
    return repoMetadata;
  }

  private async doGetSingleRepo(repoId: Repo['id']): Promise<Repo> {
    const repo = await this.prismaClient.repo.findUnique({
      where: {
        id: repoId,
      },
      rejectOnNotFound: notFoundFactory(RepoNotFoundError),
    })

    return RepoServiceImpl.serialize(repo);
  }

  async getSingleRepoById(repoId: Repo['id']): Promise<Repo> {
    return this.doGetSingleRepo(repoId);
  }

  private async doGetSingleRepoByOwnerAndName(
    ownerType: Repo['ownerType'],
    ownerId: Repo['ownerId'],
    repoName: Repo['name']
  ): Promise<Repo> {
    const repo = await this.prismaClient.repo.findFirst({
      where: {
        ownerId: Uuid.from(ownerId),
        ownerType,
        name: repoName,
      },
      rejectOnNotFound: notFoundFactory(RepoNotFoundError),
    })

    return RepoServiceImpl.serialize(repo);
  }

  async getSingleRepoByOwnerAndName(
    ownerType: Repo['ownerType'],
    ownerId: Repo['ownerId'],
    repoName: Repo['name']
  ): Promise<Repo> {
    return this.doGetSingleRepoByOwnerAndName(ownerType, ownerId, repoName);
  }

  private async doDeleteRepo(repoName: string, ownerName: string, ownerType: OwnerType): Promise<void> {
    try {
      const ownerTypeDirName = this.ownerService.getOwnerTypeDirName(ownerType);
      const repoBasePath = RepoServiceImpl.getRepoBasePath(repoName, ownerName, ownerTypeDirName);
      await this.gitService.delete(repoBasePath);
    } catch (causeRaw) {
      const cause = causeRaw as Error;
      throw new UnableToDeleteRepoError('Something went wrong while deleting the repo.', { cause, });
    }
  }

  async deleteRepo(repoId: Repo['id'], userRaw?: User): Promise<void> {
    const user = RepoServiceImpl.authenticate(userRaw);
    if (!user) {
      throw new UnauthorizedToDeleteRepoError('Could not delete repo with insufficient authorization.');
    }

    const repo = await this.doGetSingleRepo(repoId);
    const ownerName = await this.ownerService.getOwnerName(repo.ownerId, repo.ownerType);
    await this.doDeleteRepo(repo.name, ownerName, repo.ownerType as OwnerType);
    await this.repoLogService.logDeleteRepo(user, repo);
  }

  private async doGetOwnedRepos(ownerType: Repo['ownerType'], ownerId: Repo['ownerId']) {
    const repos = await this.prismaClient.repo.findMany({
      where: {
        ownerId,
        ownerType,
      },
    })

    return repos.map((r) => ({
      ...r,
      id: Uuid.from(r.id),
      ownerId: Uuid.from(r.ownerId),
    }));
  }

  async getOwnedRepos(ownerType: Repo['ownerType'], ownerId: Repo['ownerId']): Promise<Repo[]> {
    return this.doGetOwnedRepos(ownerType, ownerId);
  }

  private static gitPackSideband(s: string): string {
    return `${(4 + s.length).toString(16).padStart(4, '0')}${s}`;
  }

  async doGetRefsFromService(
    ownerType: Repo['ownerType'],
    ownerName: Repo['name'],
    repoName: Repo['name'],
    service?: string,
  ) {
    if (!service) {
      throw new UnknownServiceError('Could not return refs for service.');
    }

    const gitServiceMethods: Record<RepoAvailableService, GitServiceMethod> = {
      [RepoAvailableService.RECEIVE_PACK]: this.gitService.advertiseReceivePackRefs,
      [RepoAvailableService.UPLOAD_PACK]: this.gitService.advertiseUploadPackRefs,
    }
    const { [service as RepoAvailableService]: childProcessMethod } = gitServiceMethods;

    if (!childProcessMethod) {
      throw new UnknownServiceError('Could not return refs for service.');
    }

    const repoBasePath = RepoServiceImpl.getRepoBasePath(ownerType, ownerName, repoName);

    return new Promise<Stream>(async (resolve, reject) => {
      const stream = new PassThrough({
        emitClose: true,
        autoDestroy: true,
      });

      const childProcess = await childProcessMethod(repoBasePath);

      childProcess.on('close', (code) => {
        if (code === 0) {
          resolve(stream);
          return;
        }

        switch (code) {
        case 128:
          reject(new RepoNotFoundError(`Repo "${repoBasePath}" does not exist. Check that the repo has been created and not deleted previously.`));
          return;
        default:
          break;
        }

        reject(new UnableToInvokeRepoServiceError(`Something went wrong while invoking repo service: Git process received error code ${code}.`));
      });

      stream.push(`${RepoServiceImpl.gitPackSideband(`# service=${service}\n`)}0000`);
      if (childProcess.stdout) {
        childProcess.stdout.pipe(stream);
      }
    });
  }

  async getRefsFromService(
    ownerType: Repo['ownerType'],
    ownerName: Repo['name'],
    repoName: Repo['name'],
    service?: string,
    userRaw?: User,
  ): Promise<Stream> {
    const user = RepoServiceImpl.authenticate(userRaw);
    if (!user) {
      throw new UnauthorizedToInvokeRepoServiceError('Could not invoke repo service with insufficient authorization.');
    }

    // TODO compute user abilities here

    return this.doGetRefsFromService(ownerType, ownerName, repoName, service);
  }
}
